import itertools
from typing import Union

class DataCombination:
    _items: list
    
    def __init__(self, items):
        self._items = items

    def keys(self):
        ids = [range(len(item)) for item in self._items]
        for keys in itertools.product(*ids):
            yield keys
            
    def values(self):
        for values in itertools.product(*self._items):
            yield values

    def items(self):
        ids = [range(len(item)) for item in self._items]
        iditems = range(len(self._items))
        for keys in itertools.product(*ids):
            yield keys, tuple([self._items[i][keys[i]] for i in iditems])



from typing import Union
import copy

from .operators import *
from .expression import Expression
from .tokens import Tokens

class ExpressionSolver:

    tokens: list
    operators: dict
    expr: Expression
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass
    
    def __init__(self, atom, operators:dict = None, steps:list = None):
        self.tokens = Tokens(atom)
        self.operators = operators if operators else {
            'log':OperatorLog, 'log10':OperatorLog10, 'logb':OperatorLogb,
            'exp':OperatorExp, 'sqrt':OperatorSqrt,   'powb':OperatorPowb,
            'sin':OperatorSin, 'cos':OperatorCos,     'tan':OperatorTan,
            'par':OperatorPar,  # should be the last of parenthesis operators
            'pow':OperatorPow,
            'mul':OperatorMul, 'truediv':OperatorTruediv,
            'add':OperatorAdd, 'sub':OperatorSub,
            'eq':OperatorEq,   'ne':OperatorNe,
            'not':OperatorNot,  # needs to be after OperatorNe
            'le':OperatorLe,   'ge':OperatorGe,
            'lt':OperatorLt,   'gt':OperatorGt,
            'and':OperatorAnd, 'or':OperatorOr, 
        }
        self.steps = steps if steps else [
            dict(operators=['log', 'log10', 'logb', 'exp', 'sqrt', 'powb', 'sin', 'cos', 'tan', 'par'], otype=Otype.ARGS),
            dict(operators=['add', 'sub'],     otype=Otype.UNARY),
            dict(operators=['pow'],            otype=Otype.BINARY),
            dict(operators=['mul', 'truediv'], otype=Otype.BINARY),
            dict(operators=['add', 'sub'],     otype=Otype.BINARY),
            dict(operators=['eq', 'ne', 'le', 'ge', 'lt', 'gt'], otype=Otype.BINARY),
            dict(operators=['not'],            otype=Otype.UNARY),
            dict(operators=['and'],            otype=Otype.BINARY),
            dict(operators=['or'],             otype=Otype.BINARY),
        ]

    def solve(self, expr:Union[str,Expression]):
        self.expr = Expression(expr) if isinstance(expr, str) else expr
        
        # Tokenize expression
        while self.expr.right:
            for operator in self.operators.values():
                if self.expr.right.startswith(operator.symbol):
                    # Create atom from the left side and append it to tokens
                    if left:=self.expr.pop_left():
                        self.tokens.append(self.tokens.atom(left))
                    # Initialize an operator
                    op = operator(self.expr)
                    if op.args:
                        #print(op.args, steps)
                        # Solve operator arguments
                        with ExpressionSolver(self.tokens.atom, self.operators, self.steps) as es:
                            for a in range(len(op.args)):
                                op.args[a] = es.solve(op.args[a])
                    # Append operator to tokens
                    self.tokens.append(op)
                    break
            else:
                self.expr.shift()
        # Create atom from the remaining left side
        if left:=self.expr.pop_left():
            self.tokens.append(self.tokens.atom(left))

        # Perform operation steps
        for o,ostep in enumerate(list(self.steps)):
            operators = tuple([self.operators[o] for o in ostep['operators'] if o in self.operators.keys()])
            if operators:
                self.tokens.operate(operators, ostep['otype'])

        # Return the final atom
        if len(self.tokens.left)>0 or len(self.tokens.right)>1:
            raise Exception("Cannot solve expression due to unprocessed tokens:", self.tokens.left, self.tokens.right)
        atom = self.tokens.get_right()
        #print(self.expr.expr, atom)
        return atom



from .node_base import BaseNode
from .parser import Parser
from ..settings import Namespace
from ...units import Quantity, UnitEnvironment

class UnitNode(BaseNode):
    keyword: str = 'unit'

    @staticmethod
    def is_node(parser):
        parser.kwd_unit()
        if parser.is_parsed('kwd_unit'):
            parser.part_comment()
            return UnitNode(parser)
    
    def parse(self, env):
        parser = Parser(
            code=self.value_raw,
            source=self.source
        )
        parser.part_reference()
        if parser.is_parsed('part_reference'):
            # import a remote source
            units = env.request(parser.value_ref, namespace=Namespace.UNITS)
            for key,val in units.items():
                env.add_unit(key, val.unit)
        else:
            parser.part_name(path=False) # parse name
            parser.part_equal()          # parse equal sign
            parser.part_value()          # parse value
            if parser.value_ref:
                # inject value of a node
                self.inject_value(env, parser)
            parser.part_units()          # parse unit
            with UnitEnvironment(env.units):
                unit = Quantity(float(parser.value_raw), parser.units_raw)
                unit.symbol = '['+parser.name+']'
                unit.dfn = f"{parser.value_raw}*{parser.units_raw}"
                env.add_unit(parser.name, unit)
        return None



from .node_base import BaseNode
from ..settings import Sign, Keyword
from ..solvers import LogicalSolver

class CaseNode(BaseNode):
    keyword: str = 'case'

    @staticmethod
    def is_node(parser):
        parser.kwd_case()
        if parser.is_parsed('kwd_case'):
            if parser.name.endswith(Sign.CONDITION + Keyword.CASE):
                parser.part_value()
            parser.part_comment()
            return CaseNode(parser)
            
    def parse(self, env):
        # Solve case
        if self.name.endswith(Sign.CONDITION + Keyword.CASE):
            with LogicalSolver(env) as s:
                if self.value_expr:
                    self.value = s.solve(self.value_expr)
                else:
                    self.value = s.solve(self.value_raw)
        return None



from .settings import *
from .quantity import Quantity

def check_unique_symbols():   
    units = list(UNIT_STANDARD.keys())
    for symbol, unit in UNIT_STANDARD.items():
        if isinstance(unit.prefixes, list):
            for prefix in unit.prefixes:
                assert prefix in UNIT_PREFIXES
                units.append(f"{prefix}{symbol}")
        elif unit.prefixes is True:
            for prefix in UNIT_PREFIXES.keys():
                units.append(f"{prefix}{symbol}")
    #print(units)
    #exit(1)
    units.sort()
    seen = set()
    dupes = [x for x in units if x in seen or seen.add(x)]    
    if len(dupes)==0:
        return True
    else:
        raise Exception("Following unit symbols are duplicated:", dupes)
    
class UnitEnvironment:
    
    new_units: list
    new_types: list
    
    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()
    
    def __init__(self, units):
        self.new_units = []
        self.new_types = []
        for symbol, unit in units.items():
            if isinstance(unit, Quantity):
                unit = {'magnitude':unit.magnitude.value*unit.baseunits.magnitude, 'dimensions':unit.baseunits.dimensions.value(dtype=list)}
            if symbol in UNIT_STANDARD:
                raise Exception("Unit with this symbol already exists:", symbol)
            if 'name' not in unit:
                unit['name'] = symbol
            if 'definition' not in unit:
                unit['definition']=None
            elif unit['definition'] is not None:
                if not isinstance(unit['definition'], str) and unit['definition'] not in UNIT_TYPES:
                    UNIT_TYPES.insert(0, unit['definition'])
                    self.new_types.append(unit['definition'])
            if 'prefixes' not in unit:
                unit['prefixes'] = False
            UNIT_STANDARD.append(symbol, (unit['magnitude'], unit['dimensions'], unit['definition'], unit['name'], unit['prefixes']))
            self.new_units.append(symbol)
        check_unique_symbols()
        
    def close(self):
        for unit in self.new_units:
            del UNIT_STANDARD[unit]
        for utype in self.new_types:
            UNIT_TYPES.remove(utype)
        
        
    


import os
import sphinx
import pathlib
from docutils import nodes
from docutils.parsers.rst import Directive, directives
from inspect import getframeinfo, stack
import collections

from ..dip import DIP
from ..source import Source
from ..settings import Format
from ..datatypes import NumberType
from ..nodes import IntegerNode, StringNode, FloatNode

class SphinxDocs(Directive):

    has_content = True

    option_spec = {
        'show-code': directives.flag,
    }
    
    def run(self):

        # We are calling this script from a Sphinx base directory,
        # so DIP file paths are defined relative to it.
        source = Source(lineno=1, filename=os.getcwd())
        file_dip = self.content[0]
        
        with DIP(source=source, docs=True) as dip:
            dip.from_file(file_dip)
            env = dip.parse()
            sources = env.sources
            units = env.units
            data = env.data(format=Format.NODE)

        data = collections.OrderedDict(sorted(data.items()))

        def add_field(key, value, literal=False, options=False):
            field = nodes.field()
            field += nodes.field_name('',key)
            if literal:
                field += nodes.field_body('',nodes.literal(text=value))
            elif options:
                opts = []
                for v in value:
                    if v.units_raw:
                        opts.append(f"{v.value_raw} {v.units_raw}")
                    else:
                        opts.append(f"{v.value_raw}")
                field += nodes.field_body('',nodes.inline('',", ".join(opts))) #blist)
            else:
                field += nodes.field_body('',nodes.inline(text=value))
            return field

        nodelist = []
        
        section = nodes.section(ids=["dip-sources"])
        section += nodes.title('','Sources')
        if sources:
            deflist = nodes.definition_list()
            for name, source in sources.items():
                props = nodes.field_list()
                props += add_field('File', source.path)
                defitem = nodes.definition_list_item()
                defitem += nodes.term('', '', nodes.strong(text=name))
                defitem += nodes.definition('', props)
                deflist += defitem
            section += deflist
        else:
            section += nodes.paragraph('','There are no reference sources in this DIP file')
        nodelist.append(section)
        
        section = nodes.section(ids=["dip-units"])
        section += nodes.title('','Units')
        if units:
            deflist = nodes.definition_list()
            for name, unit in units.items():
                props = nodes.field_list()
                props += add_field('Definition', f"{unit['magnitude']} {unit['name']}")
                defitem = nodes.definition_list_item()
                defitem += nodes.term('', '', nodes.strong(text=unit['name']))
                defitem += nodes.definition('', props)
                deflist += defitem
            section += deflist
        else:
            section += nodes.paragraph('','There are no custom units in this DIP file')
        nodelist.append(section)
        
        section = nodes.section(ids=["dip-parameters"])
        section += nodes.title('','Parameters')
        deflist = nodes.definition_list()
        for name,node in data.items():
            text =  name+"\n"
            props = nodes.field_list()
            if 'show-code' in self.options:
                props += add_field("Line", node.source.lineno)
            if node.defined:
                props += add_field("Node type", 'declaration')
            else:
                if node.constant:
                    props += add_field("Node type", 'constant definition')
                else:
                    props += add_field("Node type", 'definition')
            props += add_field("Data type", node.keyword)
            if node.value:
                props += add_field("Default value", node.value.value, literal=True)
            if node.units_raw:
                props += add_field("Unit", node.units_raw)
            if node.condition:
                props += add_field("Condition", node.condition, literal=True)
            if isinstance(node,(IntegerNode,FloatNode,StringNode)):
                props += add_field("Options", node.options, options=True)
            defitem = nodes.definition_list_item()
            defitem += nodes.term('', '', nodes.strong(text=text))
            defitem += nodes.definition('', props)
            deflist += defitem
        section += deflist
        nodelist.append(section)
        

        if 'show-code' in self.options:
            section = nodes.section(ids=["dip-code"])
            section += nodes.title('','Code')
            section += sphinx.addnodes.highlightlang('', nodes.paragraph(text='width float = 23 cm'), lang='DIP', force=False, linenothreshold=-1)
            with open(file_dip) as f:
                text = f.read()
                text = text.replace("\t"," "*8)
                block = nodes.literal_block(text=text)
                section += block
            nodelist.append(section)
        
        return nodelist


def setup(app):
    app.add_directive("dipdocs", SphinxDocs)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from .node_base import BaseNode
from .node_select import SelectNode
from ..solvers import TemplateSolver, FunctionSolver
from ..datatypes import StringType

class StringNode(BaseNode, SelectNode):
    keyword: str = 'str'
    format: str = None

    @staticmethod
    def is_node(parser):
        if parser.keyword=='str':
             parser.part_dimension()
             parser.part_equal()
             if parser.is_parsed('part_equal'): # definition
                 parser.part_value()  
             else:
                 parser.defined = True  # declaration
             parser.part_units()    
             parser.part_comment()
             return StringNode(parser)
         
    def set_value(self, value=None):
        """ Set value using value_raw or arbitrary value
        """
        if value is None and self.value_raw:
            self.value = StringType(self.cast_value())
        elif value:
            self.value = StringType(value)
        else:
            self.value = None
            
    def parse(self, env):
        if self.value_fn: # Process function
            with FunctionSolver(env) as s:
                self.value_raw = s.solve(self.value_fn)
        if self.value_expr: # Process function
            with TemplateSolver(env, source=self.source) as s:
                self.value_raw = s.solve(self.value_expr)
        if self.units_raw:
            raise Exception('String datatype does not support units:', self.code)
        return None    



from .node_base import BaseNode

class GroupNode(BaseNode):
    keyword: str = 'group'

    @staticmethod
    def is_node(parser):
        parser.part_comment()
        if parser.is_empty():
            return GroupNode(parser)



class Expression:
    
    left: str
    expr: str
    right: str
    
    def __repr__(self):
        return f"Expr({self.right.strip()})"
    
    def __init__(self, expr:str):
        self.expr = expr
        self.right = expr
        self.left = ''
        
    def shift(self, nchar:int=1):
        self.left += self.right[:nchar]
        self.right = self.right[nchar:]
        
    def remove(self, string:str):
        self.right = self.right[len(string):]
        
    def pop_left(self):
        left = self.left.strip()
        self.left = ''
        return left



import numpy as np
import os
import pytest
import sys
sys.path.insert(0, 'src')

from scinumtools.solver import *

def test_custom_atom1():

    # test with external variables
    foo = 3
    bar = 4
    
    class Atom(AtomBase):
        
        def __init__(self, value:str):
            if value=='foo':
                self.value = foo
            elif value=='bar':
                self.value = bar
            else:
                self.value = float(value)
                
    expressions = {
        'foo < bar && foo * bar == 12':    foo < bar and foo*bar==12,
        'foo*bar**(2-233)':                foo*bar**(2-233),
        'sin(foo)':                        np.sin(foo),
    }
    with ExpressionSolver(Atom) as es:
        for expr, value in expressions.items():
            result = es.solve(expr)
            assert result.value == value
        
def test_custom_atom2():
    
    # modify atom to operate with strings
    class AtomCustom(AtomBase):
        value: str
        def __init__(self, value:str):
            self.value = str(value)
        def __add__(self, other):
            return AtomCustom(self.value + other.value)
        def __gt__(self, other):
            return AtomCustom(len(self.value) > len(other.value))
    operators = {'add':OperatorAdd,'gt':OperatorGt}
    # test simpple functions
    with ExpressionSolver(AtomCustom, operators) as es:
        result = es.solve('foo + bar')
        assert result.value == 'foobar'
        
    steps = [
        dict(operators=['add'],  otype=Otype.BINARY),
        dict(operators=['gt'],   otype=Otype.BINARY),
    ]
    # test multiple functions
    with ExpressionSolver(AtomCustom, operators, steps) as es:
        result = es.solve("limit + 100 km/s > limit + 50000000000 km/s")
        assert result.value == 'False' # AtomCustom returns strings
    
def test_operator_selection():

    expressions = {
        '23 > 4':    23 > 4,
        '20 == 20':  20. == 20.,
    }
    
    # specify operators
    operators = {'gt':OperatorGt,'eq':OperatorEq}
    with ExpressionSolver(AtomBase, operators) as es:
        for expr, value in expressions.items():
            result = es.solve(expr)
            assert result.value == value

    # throw error if operator is not specified
    operators = {'log':OperatorLog}
    with ExpressionSolver(AtomBase, operators) as es:
        for expr, value in expressions.items():
            with pytest.raises(ValueError) as excinfo:
                result = es.solve(expr)
            assert str(excinfo.value)==f"could not convert string to float: '{expr}'"

def test_operator_modification():

    # modify existing operators
    class CustomOperatorNot(OperatorNot):
        symbol: str = 'not'
    operators = {'not':CustomOperatorNot}
    with ExpressionSolver(AtomBase, operators) as es:
        result = es.solve('not 1')
        assert result.value == False
    
def test_custom_operator():
    
    # create a new operator
    class OperatorSquare(OperatorBase):   # operate from left side
        symbol: str = '~'
        def operate_unary(self, tokens):
            right = tokens.get_right()
            tokens.put_left(right*right)
    class OperatorCube(OperatorBase):     # operate from right side
        symbol: str = '^'
        def operate_unary(self, tokens):
            left = tokens.get_left()
            tokens.put_left(left*left*left)
    operators = {'square':OperatorSquare,'cube':OperatorCube,'add':OperatorAdd}
    steps = [
        dict(operators=['square','cube'], otype=Otype.UNARY),
        dict(operators=['add'],           otype=Otype.BINARY),
    ]
    with ExpressionSolver(AtomBase, operators, steps) as es:
        result = es.solve('~3 + 2^')
        assert result.value == 17
        
def test_customisation_in_parentheses():
    
    # modify atom to operate with strings
    class AtomCustom(AtomBase):
        value: str
        def __init__(self, value:str):
            self.value = str(value)
        def __add__(self, other):
            return AtomCustom(self.value + other.value)
        def __gt__(self, other):
            return AtomCustom(len(self.value) > len(other.value))
    operators = {'add':OperatorAdd,'gt':OperatorGt,'par':OperatorPar}
    steps = [
        dict(operators=['par'],  otype=Otype.ARGS),
        dict(operators=['add'],  otype=Otype.BINARY),
        dict(operators=['gt'],   otype=Otype.BINARY),
    ]
    with ExpressionSolver(AtomCustom, operators, steps) as es:
        # test multiple functions
        result = es.solve("(limit + 100 km/s) > (limit + 50000000000 km/s)")
        assert result.value == 'False' # AtomCustom returns strings


import numpy as np
from enum import Enum

from .expression import Expression

class Otype(Enum):
    
    ARGS = 0
    UNARY = 1
    BINARY = 2
    TERNARY = 3
    
class OperatorBase:
    
    symbol: str       # Operator symbol
    args: list = None # Arguments of a parenthesis operator
    
    def __init__(self, expr:str=None):
        if expr:
            expr.remove(self.symbol)
            
    def __repr__(self):
        return f"Oper({self.symbol})"
    
class OperatorAdd(OperatorBase):
    
    symbol: str = '+'
    
    def operate_unary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        if left is None and isinstance(right, tokens.atom):
            tokens.put_left(right)        
        elif isinstance(right, OperatorAdd):
            tokens.put_left(left)
            tokens.put_right(right)
        elif isinstance(right, OperatorSub):
            tokens.put_left(left)
            tokens.put_right(right)
        elif not isinstance(left, tokens.atom) and isinstance(right, tokens.atom):
            tokens.put_left(left)
            tokens.put_right(right)
        else:
            tokens.put_left(left)
            tokens.put_left(OperatorAdd())
            tokens.put_right(right)
            
    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left + right)

class OperatorSub(OperatorBase):
    
    symbol: str = '-'
    
    def operate_unary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        if left is None and isinstance(right, tokens.atom):
            tokens.put_left(-right)
        elif isinstance(right, OperatorAdd):
            tokens.put_left(left)
            tokens.put_right(OperatorSub())
        elif isinstance(right, OperatorSub):
            tokens.put_left(left)
            tokens.put_right(OperatorAdd())
        elif not isinstance(left, tokens.atom) and isinstance(right, tokens.atom):
            tokens.put_left(left)
            tokens.put_right(-right)
        else:
            tokens.put_left(left)
            tokens.put_left(OperatorSub())
            tokens.put_right(right)
            
    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left - right)

class OperatorMul(OperatorBase):
    
    symbol: str = '*'
    
    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left * right)

class OperatorTruediv(OperatorBase):
    
    symbol: str = '/'
    
    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left / right)

class OperatorPow(OperatorBase):

    symbol: str = '**'

    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left ** right)
        
class OperatorPar(OperatorBase):
    
    symbol: str = '('            # operator specifc opening symbol
    symbol_open: str = '('       # general opening symbol
    symbol_separator: str = ','  # argument separator
    symbol_close: str = ')'      # general closing symbol 
    narg: int = 1
    
    def __init__(self, expr:str):
        super().__init__(expr)
        depth=1
        self.args = []
        while depth>0:
            if len(expr.right)==0:
                raise Exception("Unclosed parenthesis in", expr.expr)
            elif expr.right.startswith(self.symbol) or expr.right.startswith(self.symbol_open):
                depth += 1
            elif expr.right.startswith(self.symbol_separator) and depth==1:
                expr.remove(self.symbol_separator)
                self.args.append(Expression(expr.pop_left()))                
            elif expr.right.startswith(self.symbol_close):
                depth -= 1
                if depth==0:
                    expr.remove(self.symbol_close)
                    self.args.append(Expression(expr.pop_left()))
                    break
            expr.shift()
        if len(self.args)!=self.narg:
            raise Exception("Wrong number of arguments:", expr.expr, len(self.args), self.narg)
        
    def operate_args(self, tokens):
        tokens.put_left(self.args[0])

class OperatorExp(OperatorPar):

    symbol: str = 'exp('

    def operate_args(self, tokens):
        tokens.put_left(tokens.atom(np.e)**self.args[0])        
            
class OperatorLog(OperatorPar):

    symbol: str = 'log('

    def operate_args(self, tokens):
        tokens.put_left(self.args[0].log())        

class OperatorLog10(OperatorPar):

    symbol: str = 'log10('

    def operate_args(self, tokens):
        tokens.put_left(self.args[0].log10())        

class OperatorSqrt(OperatorPar):

    symbol: str = 'sqrt('

    def operate_args(self, tokens):
        tokens.put_left(self.args[0].sqrt())        

class OperatorSin(OperatorPar):

    symbol: str = 'sin('

    def operate_args(self, tokens):
        tokens.put_left(self.args[0].sin())        

class OperatorCos(OperatorPar):

    symbol: str = 'cos('

    def operate_args(self, tokens):
        tokens.put_left(self.args[0].cos())        

class OperatorTan(OperatorPar):

    symbol: str = 'tan('

    def operate_args(self, tokens):
        tokens.put_left(self.args[0].tan())        

class OperatorLogb(OperatorPar):

    symbol: str = 'logb('
    narg: int = 2

    def operate_args(self, tokens):
        tokens.put_left(self.args[0].log()/self.args[1].log())        

class OperatorPowb(OperatorPar):

    symbol: str = 'pow('
    narg: int = 2

    def operate_args(self, tokens):
        tokens.put_left(self.args[0]**self.args[1])     
        
class OperatorAnd(OperatorBase):
    
    symbol: str = '&&'
    
    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left.logical_and(right))

class OperatorOr(OperatorBase):
    
    symbol: str = '||'
    
    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left.logical_or(right))

class OperatorNot(OperatorBase):
    
    symbol: str = '!'
    
    def operate_unary(self, tokens):
        right = tokens.get_right()
        tokens.put_right(right.logical_not())
        
class OperatorEq(OperatorBase):

    symbol: str = '=='

    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left == right)

class OperatorNe(OperatorBase):

    symbol: str = '!='

    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left != right)

class OperatorLe(OperatorBase):

    symbol: str = '<='

    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left <= right)

class OperatorGe(OperatorBase):

    symbol: str = '>='

    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left >= right)

class OperatorLt(OperatorBase):

    symbol: str = '<'

    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left < right)

class OperatorGt(OperatorBase):

    symbol: str = '>'

    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        tokens.put_left(left > right)



import numpy as np

from .quantity import Quantity

class NaN:

    def __new__(cls, unit=None):
        if unit:
            return Quantity(np.nan,str(unit))
        else:
            return Quantity(np.nan)
    


import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.solvers import TemplateSolver

def test_formatting():
    with DIP() as dip:
        dip.from_string("""
id int = 345
name str = 'Tina'
body
  weight float = 62.3 kg
  height float = 177 cm
married bool = true
        """)
        env = dip.parse()
    with TemplateSolver(env) as p:
        assert p.solve("""
ID:      {{?id}:05d}
Name:    {{?name}}
Weight:  {{?body.weight}:.3e}
Height:  {{?body.height}:.2f}
Married: {{?married}}
        """) == """
ID:      00345
Name:    Tina
Weight:  6.230e+01
Height:  177.00
Married: True
        """

def test_arrays():
    with DIP() as dip:
        dip.from_string("""
name str = "Will Smith"
widths float[2,3] = [[23.4,235.4,34],[1e10,2e23,5e20]]
        """)
        env = dip.parse()
    with TemplateSolver(env) as p:
        assert p.solve("""
Surname:  {{?name}[5:]}
Scalar:   {{?widths}[1,1]:.2e}
Array:
{{?widths}[:,1:]}
        """) == """
Surname:  Smith
Scalar:   2.00e+23
Array:
[[2.354e+02 3.400e+01]
 [2.000e+23 5.000e+20]]
        """

def test_strings():
    with DIP() as dip:
        dip.from_string("""
name str = "William Smith"
        """)
        env = dip.parse()
    with TemplateSolver(env) as p:
        assert p.solve("""
Surname: {{?name}[8:]}
        """) == """
Surname: Smith
        """
        
if __name__ == "__main__":
    # Specify wich test to run
    test = sys.argv[1] if len(sys.argv)>1 else True

    # Loop through all tests
    for fn in dir(sys.modules[__name__]):
        if fn[:5]=='test_' and (test is True or test==fn[5:]):
            print(f"\nTesting: {fn}\n")
            locals()[fn]()



from .node_base import BaseNode
from ..solvers import LogicalSolver, FunctionSolver
from ..datatypes import BooleanType

class BooleanNode(BaseNode):
    keyword: str = 'bool'
    value: bool = None
    dtype = bool

    @staticmethod
    def is_node(parser):
        if parser.keyword=='bool':
             parser.part_dimension()
             parser.part_equal()
             if parser.is_parsed('part_equal'): # definition
                 parser.part_value()  
             else:
                 parser.defined = True  # declaration
             parser.part_units()    
             parser.part_comment()
             return BooleanNode(parser)
    
    def set_value(self, value=None):
        """ Set value using value_raw or arbitrary value
        """
        if value is None:  # == None
            if self.value_raw:
                self.value = BooleanType(self.cast_value())
            else:
                self.value = None
        else:              # == True/False
            self.value = BooleanType(value)
            
    def parse(self, env):
        if self.value_fn: # Process function
             with FunctionSolver(env) as s:
                self.value_raw = s.solve(self.value_fn)
        if self.value_expr: # Process expression
            with LogicalSolver(env) as s:
                self.value_raw = s.solve(self.value_expr)
        if self.units_raw:
            raise Exception('Boolean datatype does not support units:', self.code)
        return None    



import numpy as np
import os
import sys
sys.path.insert(0, 'src')

import scinumtools as snt

def test_caching():

    file_cache = "tests/cached_data.npy"
    
    if os.path.isfile(file_cache):
        os.remove(file_cache)
    assert not os.path.isfile(file_cache)
    
    @snt.CachedFunction(file_cache)
    def read_data(a, b):
        return dict(a=a, b=b)

    data = read_data('foo','bar')    
    assert data == dict(a='foo', b='bar')

    data = read_data('foo2','bar2')
    assert os.path.isfile(file_cache)
    assert data == dict(a='foo', b='bar')
    
    if os.path.isfile(file_cache):
        os.remove(file_cache)
    assert not os.path.isfile(file_cache)

    @snt.CachedFunction(file_cache)
    def read_data():
        return {
            'array': np.linspace(1,100,100),
            'scalar': 34,
            'string': 'foo',
        }
    
    data = read_data()
    np.testing.assert_equal(data, dict(
        array=np.linspace(1,100,100),
        scalar=34,
        string='foo',
    ))



from dataclasses import dataclass
import numpy as np
import copy

@dataclass
class Type:
    value: str = None
    unit: str = None

    def copy(self):
        """ Copy a new object from self
        """
        return copy.copy(self)


from pygments.style import Style
from pygments.token import Token

class StyleLexer(Style):

    styles = {
        # node components
        Token.Indent:                 '#BDBDBD',        # indent tag
        Token.Keyword:                '#757575',        # directives $ and !
        Token.Name:                   'bold #616161',   # node name
        Token.Type:                   '#616161',        # node type
        Token.Dimension:              '#9E9E9E',        # node dimension
        Token.String:                 '#558B2F',        # value string
        Token.Number:                 '#B71C1C',        # value number
        Token.Boolean:                'bold #0D47A1',   # value bool
        Token.Reference:              '#E64A19',        # reference
        Token.Slice:                  '#FF8A65',        # reference slice     
        Token.Expression:             '#8D6E63',        # expression
        Token.Unit:                   '#607D8B',        # unit
        Token.Comment:                'italic #FB8C00', # comment

        # schema components
        Token.Name.Tag:               '#0097A7',        # tag

        # other components
        Token.Token.Error:            'border:#F44336', # missing highlight

        # default components
        Token.Function:               '#8D6E63',
    }

def pygments_monkeypatch_style(mod_name, cls):
    import sys
    import pygments.styles
    cls_name = cls.__name__
    mod = type(__import__("os"))(mod_name)
    setattr(mod, cls_name, cls)
    setattr(pygments.styles, mod_name, mod)
    sys.modules["pygments.styles." + mod_name] = mod
    from pygments.styles import STYLE_MAP
    STYLE_MAP[mod_name] = mod_name + "::" + cls_name



from typing import List
import re

from ..settings import Keyword, Sign
from .node import Node

class Parser(Node):
    ccode: str
    comment: str = None
    formating: str = None
    parsed: List[str]     # list of parsed directives or node parts
    
    def __init__(self, *args, **kwargs):
        self.parsed = []
        kwargs['ccode'] = kwargs['code']
        super().__init__(*args, **kwargs)
    
    def _strip(self, text):
        self.ccode = self.ccode[len(text):]

    def is_empty(self):
        return self.ccode.strip()==''

    def is_parsed(self, name:str):
        """ Check if directive or node part was already parsed

        :param str name: Name of a directive or node part
        """
        return name in self.parsed
    
    #####################
    # Directive keywords
    #####################
    
    def kwd_case(self):
        m=re.match(
            r'^(([a-zA-Z0-9_.-]*'+
            re.escape(Sign.CONDITION) + Keyword.CASE +
            r')\s+)',
            self.ccode
        )
        if m:
            self.parsed.append('kwd_case')
            self.name = m.group(2)
            self._strip(m.group(1))
        else:
            m=re.match(
                r'^([a-zA-Z0-9_.-]*('+
                re.escape(Sign.CONDITION) + Keyword.ELSE +'|'+
                re.escape(Sign.CONDITION) + Keyword.END +'))',
                self.ccode
            )
            if m:
                self.parsed.append('kwd_case')
                self.name = m.group(1)
                self._strip(m.group(1))

    def kwd_unit(self):
        m=re.match(
            r'^(([a-zA-Z0-9_.-]*'+
            re.escape(Sign.VARIABLE) + Keyword.UNIT +
            r')\s+([^#]*))',
            self.ccode
        )
        if m:
            self.parsed.append('kwd_unit')
            self.name = m.group(2)
            self.value_raw = m.group(3)
            self._strip(m.group(1))

    def kwd_source(self):
        m=re.match(
            r'^(([a-zA-Z0-9_.-]*'+
            re.escape(Sign.VARIABLE) + Keyword.SOURCE +
            r')\s+([^#]*))',
            self.ccode
        )
        if m:
            self.parsed.append('kwd_source')
            self.name = m.group(2)
            self.value_raw = m.group(3)
            self._strip(m.group(1))

    def kwd_options(self):
        m=re.match(
            r'^(('+re.escape(Sign.VALIDATION)+Keyword.OPTIONS+r')\s+)',
            self.ccode
        )
        if m:
            self.parsed.append('kwd_options')
            self.dimension = [(None,None)]
            self._strip(m.group(1))

    def kwd_constant(self):
        m=re.match(
            r'^('+re.escape(Sign.VALIDATION)+Keyword.CONSTANT+r')',
            self.ccode
        )
        if m:
            self.parsed.append('kwd_constant')
            self._strip(m.group(1))

    def kwd_format(self):
        m=re.match(
            r'^(('+re.escape(Sign.VALIDATION)+Keyword.FORMAT+r')\s*)',
            self.ccode
        )
        if m:
            self.parsed.append('kwd_format')
            self._strip(m.group(1))
            
    def kwd_condition(self):
        m=re.match(
            r'^(('+re.escape(Sign.VALIDATION)+Keyword.CONDITION+r')\s*)',
            self.ccode
        )
        if m:
            self.parsed.append('kwd_condition')
            self._strip(m.group(1))
            
    #############
    # Node parts
    #############
    
    def part_indent(self):
        m=re.match(r'^(\s*)',self.ccode)
        if m:
            self.parsed.append('part_indent')
            self.indent = len(m.group(1))
            self._strip(m.group(1))

    def part_name(self, path=True):
        if path is True:
            m=re.match(r'^([a-zA-Z0-9_.-]+)', self.ccode)  # format of node names
        else:
            m=re.match(r'^([a-zA-Z0-9_]+)', self.ccode)    # format of unit names
        if m:
            self.parsed.append('part_name')
            self.name = m.group(1)
            self._strip(m.group(1))
            if not self.is_empty() and self.ccode[0]!=' ':
                raise Exception("Name has an invalid format: "+self.code)
        else:
            raise Exception("Name has an invalid format: "+self.ccode)
        
    def part_type(self):
        types = ['bool','int','float','str','table']
        for keyword in types:
            m=re.match(r'^(\s+'+keyword+')', self.ccode)
            if m:
                self.parsed.append('part_type')
                self.keyword = keyword
                self._strip(m.group(1))
                break
        if self.keyword is None:
            raise Exception(f"Type not recognized: {self.code}")

    def _part_dimension(self):
        pattern = r'^(\[([0-9:,]+)\])'
        m=re.match(pattern, self.ccode)
        if m:
            dims = m.group(2).split(',')
            var = []
            for dim in dims:
                if ":" in dim:
                    dmin,dmax = dim.split(':')
                    var.append((
                        int(dmin) if dmin else None,
                        int(dmax) if dmax else None
                    ))
                else:
                    var.append((int(dim), int(dim)))
            self._strip(m.group(1))
            return var
        return None
        
    def part_dimension(self):
        if dim := self._part_dimension():
            self.parsed.append('part_dimension')
            self.dimension = dim
                        
    def part_equal(self):
        m=re.match(r'^(\s*=\s*)', self.ccode)
        if m:
            self.parsed.append('part_equal')
            self._strip(m.group(1))
            
    def part_value(self):
        # Parse reference
        self.part_reference(inject=True)
        if self.is_parsed('part_reference'):
            self.parsed.append('part_value')
            return
        # Parse function
        self.part_function()
        if self.is_parsed('part_function'):
            self.parsed.append('part_value')
            return
        # Parse expression
        self.part_expression()
        if self.is_parsed('part_expression'):
            self.parsed.append('part_value')
            return
        # If not block value, parse standard text value
        m=re.match(r'^(("""(.*)"""|"(.*)"|\'(.*)\'|([^# ]+)))', self.ccode)
        if m:
            self.parsed.append('part_value')
            # Reduce matches
            results = [x for x in m.groups()[1:] if x is not None]
            # Save value
            self.value_raw = results[1]
            self._strip(m.group(1))
        if self.value_raw is None:
            raise Exception("Value cannot start with an empty string:", self.code)
        
    def part_reference(self, inject=False):
        m=re.match(r'^(\s*({([^}]*)}))', self.ccode)
        if m:
            self.parsed.append('part_reference')
            self.value_ref = m.group(3)
            self.value_raw = '' 
            if not inject:
                if self.name:
                    self.name = f"{self.name}.{m.group(2)}"
                else:
                    self.name = f"{m.group(1)}"
            self._strip(m.group(1))
            self.part_slice()

    def part_slice(self):
        if dim := self._part_dimension():
            self.parsed.append('part_slice')
            self.value_slice = dim

    def part_format(self):  # template reference text format
        m=re.match(r'^(:[0-9.]*[sdfeb]+)', self.ccode)
        if m:
            self.parsed.append('part_format')
            self.formating = m.group(1)
            self._strip(m.group(1))

    def part_function(self):
        m=re.match(r'^(\(([a-zA-Z0-9_-]*)\))', self.ccode)
        if m:
            self.parsed.append('part_function')
            # Reduce matches
            self.value_raw = ''
            self.value_fn = m.group(2)
            self._strip(m.group(1))
            
    def part_expression(self):
        m=re.match(r'^(\(("""(.*)"""|"(.*)"|\'(.*)\')\))', self.ccode)
        if m:
            self.parsed.append('part_expression')
            # Reduce matches
            results = [x for x in m.groups()[1:] if x is not None]
            self.value_raw = ''
            self.value_expr = results[1]
            self._strip(m.group(1))
                    
    def part_units(self):
        # In numerical expressions following starting
        # signs +-*/ have to be explicitely excluded
        n=re.match(r'^\s+[\/*+-]+', self.ccode)
        m=re.match(r'^(\s+([^\s#=]+))', self.ccode)
        if m and not n:
            self.parsed.append('part_units')
            self.units_raw = m.group(2)
            self._strip(m.group(1))
        
    def part_comment(self):
        m=re.match(r'^(\s*#\s*(.*))$', self.ccode)
        if m:
            self.parsed.append('part_comment')
            self.comment = m.group(2)
            self._strip(m.group(1))



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.settings import Format
from scinumtools.dip.datatypes import FloatType, IntegerType, StringType

def parse(code):
    with DIP() as p:
        p.from_string(code)
        return p.parse().data(verbose=True,format=Format.TYPE)

def test_option():
    data = parse('''
coordinates int = 1
  = 1  # linear
  = 2  # cylindrical
  = 3  # spherical
    ''')
    np.testing.assert_equal(data,{
        'coordinates': IntegerType(1),
    })
    with pytest.raises(Exception) as e_info:
        parse("""
length float cm
  = 12 cm
  = 34 cm
        """)
    assert e_info.value.args[0] == "Node value must be defined:"
    with pytest.raises(Exception) as e_info:
        parse("""
deposition bool = true
  = true
  = false
        """)
    assert e_info.value.args[0] == "Node 'bool' does not support options"
    
def test_options():
    data = parse('''
size float cm
  !options [12,13,14,15,16] cm
  !options [22,23,24,25] m
size = 23 m
    ''')
    np.testing.assert_equal(data,{
        'size': FloatType(2300, 'cm')
    })
    with pytest.raises(Exception) as e_info:
        parse("""
size float cm
  !options [12,13,14,15,16] cm
size = 11
        """)
    assert e_info.value.args[0] == "Value 'FloatType(11.0 cm)' of node 'size' doesn't match with any option:"

def test_constant():
    data = parse('''
size float = 30 cm
  !constant
    ''')
    np.testing.assert_equal(data,{
        'size': FloatType(30, 'cm')
    })
    with pytest.raises(Exception) as e_info:
        parse("""
size float = 30 cm
  !constant
size = 23
        """)
    assert e_info.value.args[0] == "Node 'size' is constant and cannot be modified:"
    
def test_format():
    data = parse('''
name str = John
  !format "[a-zA-Z]+"
    ''')
    np.testing.assert_equal(data,{
        'name': StringType('John')
    })
    with pytest.raises(Exception) as e_info:
        parse("""
name str = 7-up
  !format '[a-zA-Z]+'
        """)
    assert e_info.value.args[0] == "Node value does not match the format:"
    with pytest.raises(Exception) as e_info:
        parse("""
size float = 23 cm
  !format '[a-zA-Z]+'
        """)
    assert e_info.value.args[0] == "Format can be set only to string nodes"

def test_condition():
    data = parse('''
size float = 23 cm
  !condition ('200 mm < {?} && {?} < 30 cm')
    ''')
    np.testing.assert_equal(data,{
        'size': FloatType(23, 'cm')
    })
    with pytest.raises(Exception) as e_info:
        parse("""
size float = 23 cm
  !condition ('250 mm < {?} && {?} < 30 cm')
        """)
    assert e_info.value.args[0] == "Node does not fullfil a condition:"



import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import *
    
def test_array_arithmetics():

    # Test basic arithmetics
    q = Quantity([2,3,4], 'm')
    assert str(q) == "Quantity([2. 3. 4.] m)"
    q += Quantity(2, 'm')
    assert str(q) == "Quantity([4. 5. 6.] m)"
    q -= Quantity(2, 'm')
    assert str(q) == "Quantity([2. 3. 4.] m)"
    q /= Quantity(2)
    assert str(q) == "Quantity([1.  1.5 2. ] m)"
    q *= Quantity(2)
    assert str(q) == "Quantity([2. 3. 4.] m)"
    q /= 2
    assert str(q) == "Quantity([1.  1.5 2. ] m)"
    q *= 2
    assert str(q) == "Quantity([2. 3. 4.] m)"
    q /= [1, 1.5, 2]
    assert str(q) == "Quantity([2. 2. 2.] m)"
    q *= [1, 1.5, 2]
    assert str(q) == "Quantity([2. 3. 4.] m)"
    q = q ** 10
    assert str(q) == "Quantity([1.024e+03 5.905e+04 1.049e+06] m10)"

def test_array_slicing():

    # Array slicing
    q = Quantity([1,2,3], 'm')
    assert str(q[:2])              == "Quantity([1. 2.] m)"
    assert str(q.value(dtype=int)) == "[1 2 3]"

def test_array_unit_conversion():

    # Test unit conversion on arrays
    assert str(Quantity([1,2,3], 'm').to('km')) == "Quantity([0.001 0.002 0.003] km)"
    
def test_numpy_functions():
    
    # Test numpy functions
    assert str(np.sqrt(Quantity(4, 'm2')))             == "Quantity(2.000e+00 m)"
    assert str(np.sqrt(Quantity([4, 9, 16], 'm2')))    == "Quantity([2. 3. 4.] m)"
    assert str(np.sqrt(Quantity([4, 9, 16], 'm3')))    == "Quantity([2. 3. 4.] m3:2)"
    assert str(np.sqrt(Quantity([4, 9, 16], 'm-3')))   == "Quantity([2. 3. 4.] m-3:2)"
    assert str(np.sqrt(Quantity([4, 9, 16], 'm2:3')))  == "Quantity([2. 3. 4.] m1:3)"
    assert str(np.sqrt(Quantity([4, 9, 16], 'm2:3*g3:5*s5')))  == "Quantity([2. 3. 4.] m1:3*g3:10*s5:2)"
    assert str(np.cbrt(Quantity([8, 27, 64], 'm3')))   == "Quantity([2. 3. 4.] m)"
    assert str(np.power(Quantity([2, 3, 4], 'm'),3))   == "Quantity([ 8. 27. 64.] m3)"
    assert str(np.sin(Quantity([45, 60], "deg")))      == "Quantity([0.707 0.866])"
    assert str(np.cos(Quantity([45, 60], "deg")))      == "Quantity([0.707 0.5  ])"
    assert str(np.tan(Quantity([45, 60], "deg")))      == "Quantity([1.    1.732])"
    assert str(np.arcsin(Quantity([0.3, -0.7])))       == "Quantity([ 0.305 -0.775] rad)"
    assert str(np.arccos(Quantity([0.3, -0.7])))       == "Quantity([1.266 2.346] rad)"
    assert str(np.arctan(Quantity([0.3, -0.7])))       == "Quantity([ 0.291 -0.611] rad)"
    assert str(np.linspace(0,Quantity(20,'m'),3))      == "Quantity([ 0. 10. 20.] m)"
    assert str(np.linspace(Quantity(10,'m'),Quantity(0.03,'km'),3))  == "Quantity([10. 20. 30.] m)"
    assert str(np.absolute(Quantity(-3,'m')))          == "Quantity(3.000e+00 m)"
    assert str(np.linspace(0,Quantity(23,'km'),3))     == "Quantity([ 0.  11.5 23. ] km)"
    assert str(np.logspace(1,Quantity(3,'m'),3))       == "Quantity([  10.  100. 1000.] m)"
    assert str(np.absolute(Quantity(-3,'m')))          == "Quantity(3.000e+00 m)"
    assert str(np.abs(Quantity(-3,'m')))               == "Quantity(3.000e+00 m)"
    assert str(np.round(Quantity(2.3,'m')))            == "Quantity(2.000e+00 m)"
    assert str(np.floor(Quantity(2.3,'m')))            == "Quantity(2.000e+00 m)"
    assert str(np.ceil(Quantity(2.3,'m')))             == "Quantity(3.000e+00 m)"

def test_numpy_nan():

    assert str(NaN()) == "Quantity(nan)"
    assert str(NaN('cm')) == "Quantity(nan cm)"

def test_array_operation_sides():
    
    p = Quantity([2,3,4], 'm')
    q = Quantity([5,6,7], 'm')
    assert p+q == q+p
    assert p-q == -(q-p)
    assert q*2 == 2*q
    assert p/2 == 1/(2/p)


import numpy as np
import json

from .node import Node
from ..datatypes import Type, IntegerType, FloatType, StringType, BooleanType
from ..environment import Environment
from ..settings import Keyword

class BaseNode(Node):
    
    value: str = None       # Value
    constant: bool = False
    condition: str = None
    
    def __init__(self, parser=None, *args, **kwargs):
        if parser:
            kwargs['code'] = parser.code
            kwargs['source'] = parser.source
            kwargs['indent'] = parser.indent
            kwargs['name'] = parser.name
            kwargs['value_raw'] = parser.value_raw
            kwargs['value_ref'] = parser.value_ref
            kwargs['value_slice'] = parser.value_slice
            kwargs['units_raw'] = parser.units_raw
            kwargs['dimension'] = parser.dimension
            kwargs['defined'] = parser.defined
            if parser.keyword:
                kwargs['keyword'] = parser.keyword
            if parser.value_fn:
                kwargs['value_fn'] = parser.value_fn
            if parser.value_expr:
                kwargs['value_expr'] = parser.value_expr
        super().__init__(*args, **kwargs)
    
    def parse(self, env):
        return False

    def cast_value(self, value=None):
        """ Cast (raw-)value as a datatype self, or another node
        """
        if not value:
            if self.value is None:
                value = self.value_raw
            else:
                if isinstance(self.value, (StringType, BooleanType)):
                    value = self.value.value
                else:
                    value = self.value
        if np.isscalar(value) and value in [None,'none','None']:
            value = None
        elif self.dimension or self.value_slice:
            # cast multidimensional values
            if isinstance(value, str):
                value = np.array(json.loads(value), dtype=self.dtype)
            else:
                value = np.array(value, dtype=self.dtype)
            if self.value_slice:
                value = self.slice_value(self.value_slice, value)
            if self.dimension:
                # check if dimensions are correct
                for d,dim in enumerate(self.dimension):
                    shape = value.shape[d]
                    if dim[0] is not None and shape < dim[0]:
                        raise Exception(f"Node '{self.name}' has invalid dimension: dim({d})={shape} < {dim[0]}")
                    if dim[1] is not None and shape > dim[1]:
                        raise Exception(f"Node '{self.name}' has invalid dimension: dim({d})={shape} > {dim[1]}")
            else:
                if not np.isscalar(value):
                    raise Exception("Array value set to scalar node:",self.code,value)
        else:
            # cast scalar values
            if value is not None:
                # casting string as boolean returns true always if string is non-empty
                # that's why we need to convert it expicitely
                if self.keyword=='bool':
                    if value==Keyword.TRUE:
                        value = True
                    elif value==Keyword.FALSE:
                        value = False
                    elif not isinstance(value,(BooleanType,bool,np.bool_)):
                        raise Exception("Could not convert raw value to boolean type:",value)
                else:
                    try:
                        if isinstance(value, FloatType):
                            if self.keyword=='int':
                                value = IntegerType(value.number, value.unit)
                        elif isinstance(value, IntegerType):
                            if self.keyword=='float':
                                value = FloatType(value.number, value.unit)
                        else:
                            value = self.dtype(value)
                    except:
                        raise Exception("Could not convert raw value to type:",self.code,value)
        return value

    def set_value(self, value=None):
        """ Set value using value_raw or arbitrary value
        """
        if value is None and self.value_raw:
            self.value = self.cast_value()
        elif value:
            self.value = value
        else:
            self.value = None
    
    def modify_value(self, node, env):
        """ Modify value taking value of a different node
        """
        if node.keyword!='mod' and node.dtype!=self.dtype:
            raise Exception(f"Datatype {self.dtype} of node '{self.name}' cannot be changed to {node.dtype}")
        if not self.value:  # create a dummy value if none
            self.set_value(node.value_raw)
        # copy value type modify values and units
        value = self.value.copy()
        value.value = self.cast_value(node.value_raw)
        if isinstance(value, (IntegerType, FloatType)):
            value.unit = node.units_raw
            value.convert(self.units_raw, env)
        self.set_value(value.value)

    def slice_value(self, slices, value=None):
        """ Slice part of the value

        :param list slices: List of tuples with slicing
        :param value: Value to be sliced. If none value of current node is taken.
        """
        if value is None:
            value = self.value
        if isinstance(value, Type):
            value = value.value
        smin, smax = slices.pop(0)
        if smin==smax and smin is not None:
            value = value[smin]
        elif smin!=smax:
            value = value[slice(smin,smax)]                  
        if slices:
            if smin==smax and smin is not None:
                return self.slice_value(slices.copy(), value)
            else:
                return np.array([self.slice_value(slices.copy(), val) for val in value])
        else:
            return value

    def inject_value(self, env:Environment, node=None):
        """ Inject value from a remote source

        :param env: Environment object
        :param node: Node object that should be injected
        """
        if not node:
            node = self
        if not node.value_ref:
            return
        nodes = env.request(node.value_ref, count=1)
        if isinstance(nodes, str):   # block import
            node.value_raw = nodes
        else:                        # node import
            node.value_raw = nodes[0].value_raw
            if not node.units_raw:
                node.units_raw = nodes[0].units_raw
        



import os
from pathlib import Path
import numpy as np
from inspect import getframeinfo, stack

from ..environment import Environment
from ..source import Source
from ..nodes.parser import Parser
from ..datatypes import Type

class TemplateSolver:

    env: Environment
    source: Source

    def __init__(self, env:Environment=None, **kwargs):
        if env:
            self.env = env
        else:
            self.env = Environment()
        if 'source' not in kwargs:
            # determine which file instantiate this class
            caller = getframeinfo(stack()[1][0])
            self.source = Source(lineno=caller.lineno, filename=caller.filename)
        else:
            self.source = kwargs['source']
            
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def solve(self, expr):
        out = ''
        kwargs = {'source':Source(lineno=0, filename='template'), 'keyword':'expr'}
        while len(expr)>0:
            sign, expr = expr[0], expr[1:]
            if sign=='{':
                p = Parser(code=expr, **kwargs)
                p.part_reference()
                p.part_slice()
                p.part_format()
                if p.value_ref and p.ccode[0]=='}':
                    nodes = self.env.request(p.value_ref, count=1)
                    if p.value_slice:
                        value = nodes[0].slice_value(p.value_slice)
                    else:
                        value = nodes[0].value
                    if isinstance(value, Type):
                        value = value.value
                    if p.formating:
                        form = "{0"+p.formating+"}"
                        out += form.format(value)
                    else:
                        out += str(value)
                    expr = p.ccode[1:]
                else:
                    out += sign
            else:
                out += sign
        return out

    def template(self, file_in, file_out=None):
        if not os.path.isabs(file_in):
            # set relative paths with respect to the source script
            parent = Path(self.source.filename).parent
            file_in = parent / file_in
        with open(file_in,'r') as f:
            template = f.read()
        text = self.solve(template)
        if file_out:
            if not os.path.isabs(file_out):
                # set relative paths with respect to the source script
                parent = Path(self.source.filename).parent
                file_out = parent / file_out
            with open(file_out,'w') as f:
                f.write(text)
        return text
            



from scinumtools.cached_function import CachedFunction
from scinumtools.normalize_data import NormalizeData
from scinumtools.data_plot_grid import DataPlotGrid
from scinumtools.thumbnail_image import ThumbnailImage
from scinumtools.data_combination import DataCombination
from scinumtools.stopwatch import Stopwatch
from scinumtools.row_collector import RowCollector
from scinumtools.parameter_table import ParameterTable
from scinumtools.progress_bar import ProgressBar

import scinumtools.units as units
import scinumtools.solver as solver